import { App, Configuration } from '@midwayjs/decorator';
import * as DefaultConfig from './config/config.default';
import * as koa from '@midwayjs/koa';
// 数据库
import * as typegoose from '@midwayjs/typegoose';
import * as Typegoose from '@typegoose/typegoose';
import { ResponseMiddleware } from './middleware/response.middleware';
import { DefaultErrorFilter } from './filter/default.filter';
import * as axios from '@midwayjs/axios';
import { ServiceError, FrameworkErrorEnum } from './error/service.error';
import { IMidwayContainer, Inject, MidwayConfigService } from '@midwayjs/core';
import * as upload from '@midwayjs/upload';
import { UseFile } from './entity/useFile.entity';
import * as bull from '@midwayjs/bull';


const axiosResponse = response => {
  if (response.status === 200) return response.data;
  else {
    throw new ServiceError('请求失败', FrameworkErrorEnum.REQUEST_FAULT);
  }
};
const axiosError = error => {
  return Promise.reject(error);
};
/**
 * 框架设置:
 *  引入
 *      koa:基础框架,
 *      typegoose:数据库相关,
 *      axios:自定义http请求工具,
 *      upload:文件上传,
 *      bull:任务队列
 */
@Configuration({
  namespace: 'free',
  imports: [koa, typegoose, axios, upload, bull],
  importConfigs: [
    {
      default: DefaultConfig,
    },
  ],
})
export class FreeConfiguration {
  @App()
  app: koa.Application;

  @Inject()
  configService: MidwayConfigService;
  /**修改配置,如果启用文件服务,那么就把项目文件地址表加上 */
  async onConfigLoad() {
    // 添加useFile表,将default连接中所有的文件都统计在该表中
    const allConfig = this.configService.getConfiguration();
    const useFile = allConfig.upload.use || false;
    if (!useFile) return;
    const mongoose = allConfig.mongoose;
    if (!mongoose) return;
    const dataSource = mongoose.dataSource;
    if (!dataSource) return;
    const defaultDataSource = dataSource.default;
    if (!defaultDataSource) return;
    let entities = defaultDataSource.entities;
    entities = [...entities, UseFile];
    return { mongoose: { dataSource: { default: { entities } } } };
  }

  getApiUrl() {
    const path = this.app.getConfig()?.swagger?.swaggerPath;
    const port = this.app.getConfig()?.koa?.port;
    if (path) console.log(`api文档: http://127.0.0.1:${port}${path}/index.html`);
  }
  async onReady(container: IMidwayContainer) {
    // TODO something
    this.app.getMiddleware().insertLast(ResponseMiddleware);
    this.app.useFilter([DefaultErrorFilter]);
    // typegoose设置
    Typegoose.setGlobalOptions({
      schemaOptions: {
        id: true,
        toJSON: { getters: true, virtuals: true },
        toObject: { getters: true, virtuals: true },
      },
      options: { allowMixed: Typegoose.Severity.ALLOW },
    });
    this.getApiUrl();
    // axios设置
    const httpServiceFactory = await container.getAsync(axios.HttpServiceFactory);
    const aos: object = this.app.getConfig('axios.clients');
    const keys = Object.keys(aos);
    for (const key of keys) {
      const a = httpServiceFactory.get(key);
      a.interceptors.response.use(axiosResponse, axiosError);
    }

    console.log('free frame is ready');
  }
}
