import { Provide } from '@midwayjs/decorator';
import { FrameworkErrorEnum, ServiceError } from '../error/service.error';
import { Types } from 'mongoose';
import { get } from 'lodash';
import { GetModel } from './getModel';
/**状态码 */
enum StatusCode {
  /**待处理 */
  PADDING = 'padding',
  /**处理成功 */
  SUCCESS = 'success',
  // /**处理失败 */
  // ERROR = 'error',
  ROLLBACK = 'rollback',
}
/**操作码 */
enum OperaCode {
  /**创建 */
  CREATE = 'create',
  /**修改 */
  UPDATE = 'update',
  /**删除 */
  DELETE = 'delete',
}
/**任务接口 */
interface IMission {
  /**表名 */
  model: string;
  /**操作 */
  opera: OperaCode;
  /**数据 */
  data?: any;
  /**任务状态 */
  status: StatusCode;
  /**任务操作前的数据(添加不需要) */
  backUp?: any;
  /**任务范围(添加不需要) */
  query?: object;
}
/**
 * 事务服务工具类
 */
@Provide()
export class TransactionService {
  mission: Array<IMission> = [];
  /**
   * 事务添加
   * @param modelName 表名
   * @param data 添加的数据
   */
  insert(modelName: string, data: object): Types.ObjectId {
    const _id = new Types.ObjectId();
    Object.assign(data, { _id });
    const mission: IMission = {
      model: modelName,
      opera: OperaCode.CREATE,
      data,
      status: StatusCode.PADDING,
    };
    this.mission.push(mission);
    return _id;
  }

  /**
   * 事务修改
   * @param modelName 表名
   * @param query 修改的数据范围
   * @param data 要修改为的数据
   */
  update(modelName: string, query: object, data: object) {
    const mission: IMission = {
      model: modelName,
      opera: OperaCode.UPDATE,
      data,
      status: StatusCode.PADDING,
      query,
    };
    this.mission.push(mission);
  }
  /**
   * 事务删除
   * @param modelName 表名
   * @param query 删除范围
   */
  delete(modelName: string, query: object) {
    const mission: IMission = {
      model: modelName,
      opera: OperaCode.DELETE,
      status: StatusCode.PADDING,
      query,
    };
    this.mission.push(mission);
  }

  async run() {
    // 需要备份的操作先备份
    await this.toRestoreData();
    for (const m of this.mission) {
      const { model: modelName, opera, status, data, query } = m;
      // 判断当前任务状态
      // 成功:下一步;
      if (status === StatusCode.SUCCESS) continue;
      try {
        // 先找到model
        const model = GetModel(modelName);
        // 再看opera,进行操作
        if (opera === OperaCode.CREATE) {
          await model.create(data);
        } else if (opera === OperaCode.UPDATE) {
          await model.updateMany(query, data);
        } else if (opera === OperaCode.DELETE) {
          await model.deleteMany(query);
        }
        m.status = StatusCode.SUCCESS;
      } catch (error) {
        // 发生错误后,当前步骤状态为padding,之前均为success.将success回滚
        // 更改标记,中断程序
        throw new ServiceError(error, FrameworkErrorEnum.SERVICE_FAULT);
        break;
      }
    }
  }

  /**
   * 针对每步的操作,进行数据备份,以便回滚还原
   */
  private async toRestoreData() {
    for (const m of this.mission) {
      const { model: modelName, opera, status, query } = m;
      // 只进行未处理的备份,处理过的不再备份
      if (status !== StatusCode.PADDING) continue;
      // 数据备份;创建不需要备份,没有
      if (opera === OperaCode.CREATE) continue;
      const model = GetModel(modelName);
      const backUp = await model.find(query).lean();

      m.backUp = backUp;
    }
  }
  /**
   * 回滚,将success的任务全部回滚
   */
  async rollback() {
    const needRollback = this.mission.filter(f => f.status === StatusCode.SUCCESS);
    for (const m of needRollback) {
      const { model: modelName, opera, data, backUp } = m;
      const model = GetModel(modelName);
      if (opera === OperaCode.CREATE) {
        const _id = get(data, '_id');
        await model.deleteOne({ _id });
      } else if (opera === OperaCode.UPDATE) {
        for (const bd of backUp) {
          await model.updateOne({ _id: bd._id }, bd);
        }
      } else if (opera === OperaCode.DELETE) {
        await model.insertMany(backUp);
      }
      m.status = StatusCode.ROLLBACK;
    }
  }
}
